/* /linux/drivers/block/double/dbtools.c
 *
 * Written by Jean-Marc Verbavatz <jmv@receptor.mgh.harvard.edu>
 * 8 Jan 1994
 *
 * Copyright 1994 by Jean-Marc Verbavatz. Redistribution of this file is
 * permitted under the GNU Public Licence
 */

#include <linux/config.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/locks.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/double.h>

char *db_buffer;
long *db_mapbuf;
static struct db_header db_header;
struct buffer_head **db_bh;

int dbtools_debug(void)
{
	if(memcmp(db_buffer+DB_MAXSZ, "$$$$", 4)) return 5;
	return 0;
}

/* Load bitmap for iblock n into memory */

int get_bitmap(struct dble_device *db, long n)
{
	if(db->bitmap_pos==n) return 1;		/* already loaded */
	if(db->bitmap_mod)		        /* update */
		if(!db_rw_iblock(db, WRITE, db->bitmap_pos+db->addr_bitmap, db->bitmap))
			return 0;
		else db->bitmap_mod = 0;
	/* read new one */
	if(db_rw_iblock(db, READ, n+db->addr_bitmap, db->bitmap)) {
		db->bitmap_pos = n;
		return 1;
	}
	return 0;
}


/* Load BAT for oblock n into memory */

int get_BAT(struct dble_device *db, long n)
{
	long i;


	i = n/(db->isize/db->BATsize);
	if(db->BAT_pos!=i) {
		if(db->BAT_mod) /* Flush */
			if(!db_rw_iblock(db, WRITE, db->BAT_pos+db->addr_BAT, db->BAT_block))
				return 0;
			else db->BAT_mod = 0;
/* Load appropriate BAT block into memory */
		if(!db_rw_iblock(db, READ, i+db->addr_BAT, db->BAT_block))
			return 0;
		db->BAT_pos = i;
	}
	return 1;
}

/* The driver has its own db_getblk, db_bmap functions as
 * the kernel functions getblk and bmap  cannot be called reliably
 * from within a driver.
 */

/* get real block number for regular file block. This information is
 * stored in iblocks 1 to addr_bitmap and is updated by the
 * set_fd ioctl call, when the device is mounted.
 */
long db_bmap(struct  dble_device *db, long block)
{
	int n;

	if(block<(n=db->isize/sizeof(long))) return db->bmap0[block];
	else {
		if(!db_rw_iblock(db, READ, block/n, (u_char *) db_mapbuf))
			return 0;
		return db_mapbuf[block%n];
	}
}

/* make buffer n available */
void move_queue(struct buffer_head *bh, int n)
{
	int i;

	for( i = n; i > 0; i--)
		db_bh[i] = db_bh[i-1];
	*db_bh = bh;
/* Make sure it's not still in use */
	if(bh->b_dirt) wait_on_buffer(bh);
	if(bh->b_count != 0) {
		printk("DouBle: too few bh buffers\n");
		wait_on_buffer(bh);
		if(bh->b_count != 0)
			panic("db_getblk: logical error count=%d\n", bh->b_count);
	}
	bh->b_count = 1;
}


/* get internal iblock buffer for physical block */
struct buffer_head *db_getblk(dev_t dev, long block, int size)
{
	struct buffer_head *bh;
	int i;

/* Check to see if block is already in a buffer */
	for(i = 0; i<DB_NR_BH; i++) {
		bh = db_bh[i];
		if(bh->b_dev == dev && bh->b_blocknr == block) { /* Yes */
			dble_stat.ihit++;
			move_queue(bh, i);
			return bh;
		}
	}
/* Not loaded; get buffer and clean it */
	dble_stat.inohit++;
	bh = db_bh[DB_NR_BH-1];
	move_queue(bh, DB_NR_BH - 1);

	lock_buffer(bh);
/* Now, clear bh structure */
	bh->b_dev = dev;
	bh->b_blocknr = block;
	bh->b_uptodate = 0;
	bh->b_req = 0;
	bh->b_dirt = 0;
	bh->b_size = size;
	bh->b_lock = 0;
	bh->b_wait = NULL;
	bh->b_reqnext = NULL;
	unlock_buffer(bh);
	return bh;
}

/* Read/Write iblock from physical file */
int db_rw_iblock(struct dble_device *db, int cmd, long n, u_char *s)
{
	int size, len, blksize, offset;
	long block, rblock;
	struct buffer_head *bh;

	if(n<0 || n>=db->iblocks) {
		printk("db_rw_iblock: %ld/%ld non existent block\n", n, db->iblocks);
		return 0;
	}
	if(!(blksize=db->db_inode->i_blksize))
		blksize=BLOCK_SIZE;
	if(blksize<=db->isize) {
		block=n*(db->isize/blksize);
		offset=0;
	} else {
		block=n/(blksize/db->isize);
		offset=(n%(blksize/db->isize))*db->isize;
	}
	len=db->isize;
	while(len>0) {
		if(db->db_flags & DB_BMAP) {
			rblock = db_bmap(db, block);
			if(!rblock) {
				printk("db_rw_iblock: NULL block %ld %ld\n", n, block);
				return 0;
			}
		} else rblock = block;
		bh = db_getblk(db->db_device, rblock, blksize);

		if(!bh->b_uptodate && ((cmd==READ) || offset || (len<blksize))) {
			dble_stat.rphysical++;
			ll_rw_block(READ, 1, &bh);
			if(db->debug >= 6) printk("DB_RW ll_rw READ\n");
			wait_on_buffer(bh);
			if(!bh->b_uptodate) {
				bh->b_count --;
				printk("db_rw_iblock: failure\n");
				return 0;
			}
		}
		size = blksize - offset;
		if(size>len) size=len;
		if(cmd==READ) {
			memcpy(s, bh->b_data+offset, size);
		} else
		/* only write if data changed */
			if(!bh->b_uptodate || memcmp(bh->b_data+offset, s, size)) {
				memcpy(bh->b_data+offset, s, size);
				bh->b_dirt=1;
				ll_rw_block(WRITE, 1, &bh);
				dble_stat.wphysical++;
				if(db->debug >= 6) printk("DB_RW ll_rw WRITE\n");
			}
		bh->b_count --;
		s+=size;
		len-=size;
		offset=0;
		block++;
		if(db->debug >= 5) printk("DB_RW DONE\n");
	}
	return 1;
}
	
/* find free iblock in bitmap */
u_long get_newbit(db)
struct dble_device *db;
{
	int j,k,l;
	u_long i,b;

	for(i = b = 0; ; b++) {
		if(!get_bitmap(db, b)) return 0;
		for(j = 0; j < db->isize; j++) if((k = db->bitmap[j]) != 0) {
			for(l = 0x80; l; l >>= 1, i++)
				if(k&l) {
					k &= ~l;
					db->bitmap[j] = k;
					db->bitmap_mod = 1;
					i += j<<3;
					return i;
				}
		}
		i += j<<3;
		if(i >= db->iblocks) {
			printk("get_new_bit: out of space\n");
			return 0;
		}
	}
}

/* clear free bit (iblock is being used) */
int clr_bit(struct dble_device *db, u_long n)
{
	u_long i;

	i = n/(db->isize<<3);
	n = n%(db->isize<<3);
	if(!get_bitmap(db, i)) return 0;
	db->bitmap[n>>3] &= ~(0x80>>(n&7));
	db->bitmap_mod = 1;
	return 1;
}

/* set free bit (iblock has been released) */
int set_bit(struct dble_device *db, u_long n)
{
	u_long i;

	i = n/(db->isize<<3);
	n = n%(db->isize<<3);
	if(!get_bitmap(db, i)) return 0;
	db->bitmap[n>>3] |= (0x80>>(n&7));
	db->bitmap_mod = 1;
	return 1;
}

/* return value of bit n */
int get_bit(struct dble_device *db, u_long n)
{
	u_long i;

	i = n/(db->isize<<3);
	n = n%(db->isize<<3);
	if(!get_bitmap(db, i)) return -1;
	return (db->bitmap[n>>3] & (0x80>>(n&7)));
}

/* Read one oblock into s */
int read_oblock(struct dble_device *db, long n, u_char *s)
{
	u_char *t;
	u_long *p;
	struct BAT *b;
	int i;

	if(n<0 || n>=db->oblocks) {
		printk("read_oblock: %ld/%ld non existent cluster\n", n, db->oblocks);
		return 0;
	}
	memset(db_buffer+DB_MAXSZ, '$', 4); /* For debugging */
	dble_stat.roblock++;

	if(!get_BAT(db, n)) return 0;
	b = (struct BAT *)(db->BAT_block+db->BATsize*(n%(db->isize/db->BATsize)));
	if(b->type == -1) {         /* Unused (empty) */
		memset(s, 0, db->osize);
		return 1;
	}

/* Fetch compressed data in s */
	for(i = 0, t = db_buffer, p = &(b->iblock); *p && i<db->ratio;
	    i++, p++, t += db->isize) {
		if(!db_rw_iblock(db, READ, *p, t)) return 0;
	}

/* decompress data */
	switch(b->type) {
		case 0: memcpy(s, db_buffer, db->osize);
			break;
		case 1:
			if((i = RLZW((code *)db_buffer, s, db->osize))<=0) {
				printk("Error RLZW %d\n", i);
				return 0;
			}
			break;
		default: printk("Error compression mode\n");
			return 0;
	}
	if(dbtools_debug()) {
		printk("read_oblock memory corruption\n");
		return 0;
	}

	return 1;
}

/* write one oblock to disk */
int write_oblock(struct dble_device *db, long n, u_char *s)
{
	int i,l;
	u_long *p;
	struct BAT *b;

	if(n < 0 || n >= db->oblocks) {
		printk("write_oblock: %ld/%ld non existent cluster\n", n, db->oblocks);
		return 0;
	}
	memset(db_buffer+DB_MAXSZ, '$', 4); /* For debugging */
	dble_stat.woblock++;

	if(!get_BAT(db, n)) return 0;
	b = (struct BAT *)(db->BAT_block+db->BATsize*(n%(db->isize/db->BATsize)));

/* Compress data */
	l = db->code;
	i = db->osize;
	switch(l) {
		case 0: 
			break;
		case 1: if((i=WLZW(s, (code *)db_buffer, db->osize, db->osize-db->isize))<=0
			   || i > db->osize-db->isize) { /* No compression achieved */
				if(i<0) {
					printk("Error WLZW %d\n", i);
					return 0;
				}
				l = 0;
				i = db->osize;
			} else s = db_buffer;
			break;
	}

	if(b->type != l) {	/* Change in compression method */
		b->type = l;
		db->BAT_mod = 1;
	}
	l = i/db->isize;	/* Number of iblocks */
	if(i%db->isize) l++;
	for(p=&(b->iblock), i=0; i<l; i++,p++) { /* Write one iblock */
		if(*p==0)	 /* Allocate new block */
			if(!(*p=get_newbit(db))) return 0;
			else db->BAT_mod = 1;
		if(!db_rw_iblock(db, WRITE, *p, s)) return 0;
		s+=db->isize;
	}
	for(;i<db->ratio;i++,p++) {
		if(*p) { /* Fewer iblocks used than before */
			if(!set_bit(db, *p)) return 0; /* release iblock */
			*p=0;
			db->BAT_mod = 1;
		}
		else break;
	}
	if(dbtools_debug()) {
		printk("write_oblock memory corruption\n");
		return 0;
	}

	return 1;
}

int db_flush(struct dble_device *db)
{
	int i, n, result = 0;

/* Flush bitmap */
	if(db->bitmap_mod)
		if(!db_rw_iblock(db, WRITE, db->bitmap_pos+db->addr_bitmap, db->bitmap))
			result = -1;
		else db->bitmap_mod = 0;

/* Flush BAT */
	if(db->BAT_mod)
		if(!db_rw_iblock(db, WRITE, db->BAT_pos+db->addr_BAT, db->BAT_block))
			result = -1;
		else db->BAT_mod = 0;

/* Flush buffers */
	for(i = DB_NR_BH; i-->0;)
		if(db_bh[i]->b_dirt) wait_on_buffer(db_bh[i]);
	for(i = DB_NR_BH, n = 0; i-->0;)
		if(db_bh[i]->b_count) n++;
	if(n>0) printk("db_flush: %d buffers still in use\n", n);

	return result;
}

/* Read header from disk */
int db_rheader(struct dble_device *db)
{
	long block;
	int blksize;
	struct buffer_head *bh;

	if(db->db_flags & DB_BMAP)	/* Regular file, get real address */
		block = bmap(db->db_inode, 0);
	else block = 0;
	if(!(blksize = db->db_inode->i_blksize))
		blksize = BLOCK_SIZE;
	bh = getblk(db->db_device, block, blksize);
	if(!bh->b_uptodate) ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);
	memcpy(&db_header, bh->b_data, sizeof(struct db_header));
	brelse(bh);

	if(db_header.Magic != DB_MAGIC) {
		printk("DouBle: wrong magic number\n");
		return 0;
	}
	db->isize = db_header.isize;
	db->osize = db_header.osize;
	if(db->osize>DB_MAXSZ) {
		printk("DouBle: size %d > MAX (%d)\n", db->osize, DB_MAXSZ);
		return 0;
	}
	db->iblocks = db_header.iblocks;
	db->oblocks = db_header.oblocks;
	db->addr_bitmap = db_header.addr_bitmap;
	db->addr_BAT = db_header.addr_BAT;
	db->addr_blocks = db_header.addr_blocks;
	db->BATsize = db_header.BATsize;
	db->ratio = db_header.ratio;
	db->code = db_header.code;
	db->bitmap_pos = -1;
	db->bitmap_mod = 0;
	db->BAT_pos = -1;
	db->BAT_mod = 0;
	db->debug = 0;

	db->bitmap = kmalloc(db->isize, GFP_KERNEL);
	db->BAT_block = kmalloc(db->isize, GFP_KERNEL);

	if(db->db_flags & DB_BMAP) {
		db->bmap0 = (long *) kmalloc(db->isize, GFP_KERNEL);
		if(db->bmap0 == NULL) {
			printk("DouBle: kmalloc returned NULL\n");
			return 0;
		}
		for(blksize = 0; blksize < db->isize/sizeof(long); blksize++)
			db->bmap0[blksize] = bmap(db->db_inode, (long)blksize);
	} else db->bmap0 = NULL;
	if(db->bitmap == NULL || db->BAT_block == NULL) {
		if(db->bitmap) kfree_s(db->bitmap, db->isize);
		if(db->BAT_block) kfree_s(db->BAT_block, db->isize);
		if(db->bmap0) kfree_s(db->bmap0, db->isize);
		printk("DouBle: kmalloc returned NULL\n");
		return 0;
	}

	return 1;
}
