/* /linux/drivers/block/lzw.c
 *
 * Written by Jean-Marc Verbavatz 12 Dec 1993 <jmv@receptor.mgh.harvard.edu>
 *
 * Copyright 1994 by Jean-Marc Verbavatz. Redistribution of this file is
 * permitted under the GNU Public Licence
 *
 */

/* LZW Compression and Decompression:
	Functions:
		RLZW(si, so, len)	Decoding
		code *si; compressed data
		byte *so; buffer for uncompressed data
		int  len; expected length of uncompressed data
	
		WLZW(si, so, len, maxlen)	Coding
		byte *si; raw data
		code *so; buffer for compressed data
		int  len; length of raw data
		int maxlen; maximum size of compressed data
		(so buffer must be slightly longer than maxlen)
   	Return code:
		 0 = Worthless CPU waste (No compression)
		-1 = General error
		-2 = Logical error
		-3 = Expand error
		>0 = OK/total length
*/
#include <linux/string.h>

#define LZSIZE 4096	/* Should stay < 32768 */ 
#define CLEAR 256	/* Clear Code */
/* 257 Reserved for future use */
#define START 258

typedef unsigned short code;
typedef unsigned char byte;

static code Prefix, Index;
static int bits,off;
static code *scode;
static byte *sbyte;
static unsigned int pcode, pbyte, length;

code *PTable, *NTable;
byte *CTable;

/* Write Next Code */
static int WCode(code wcode)
{
int todo;

pcode+=bits;
/* Quickly check output buffer size */
if((pcode>>3) > length) return 0; /* Insufficient compression achieved */

if((todo=bits+off-16)>=0) {
	*scode++|=wcode>>todo;
	*scode=wcode<<(16-todo);
	off=todo;
} else {
	*scode|=wcode<<(-todo);
	off+=bits;
}
return 1;
}

/* Read Next Code */
static short RCode(void)
{
code rcode; /* 15 bits maximum, with LZSIZE=32768; never negative */
int todo;

if((todo=bits+off-16)>=0) {
	rcode=(*scode++)<<todo;
	rcode|=(*scode)>>(16-todo);
	off = todo;
} else {
	rcode=(*scode)>>(-todo);
	off+=bits;
}
rcode&=(1<<bits)-1;
return (short) rcode;
}

/* Initialization for coding */
static void WInitTable(void)
{

bits=8;
pbyte=pcode=off=0;
memset(PTable, -1, LZSIZE*sizeof(code));
Index=START;
}

/* Initialization for decoding */
static void RInitTable(void)
{
int i;
byte *p;

WInitTable();
for(i=0, p=CTable; i<CLEAR; i++)
	*p++ = i;
}

/* Lookup Table */
static code LookUp(byte Car)
{
code pi;

pi=PTable[Prefix];
while(((short)pi)!=-1) {
	if(CTable[pi]==Car) return pi;
	pi=NTable[pi];
}
return -1; /* No Hit */
}

/* byte is added to table and becomes prefix */
static int WAddPrefix(void)
{
code pi;

pi=PTable[Prefix];
PTable[Prefix]=Index; /* Next entry */
NTable[Index]=pi;
CTable[Index++]=*sbyte;

if(Index==LZSIZE) { /* Table full */
	if(!WCode(CLEAR)) return 0; 
	Index=START;
	memset(PTable, -1, LZSIZE*sizeof(code));
	bits = 8;
} else if(Index>(1<<bits))
	bits++;
return 1;
}

static int Expand(code val)
{
code p,q,r;

if(val>Index) return -2;
q = -1;
PTable[Index]=Prefix;
do {
	p=val;
	while((p=PTable[(r=p)])!=q); /* Get previous */
	*sbyte++=CTable[r];
	if(++pbyte>=length) return 0; /* Done; skip useless stuff */
	if(((short)q)==-1) {
		if(Index==LZSIZE) return -3;
		CTable[Index++]=CTable[r];
	}
	q=r;
} while(q!=val);
Prefix=val;
return 1;
}

/* Write Lzw (coding) */
int WLZW(byte *si, code *so, int len, int maxlen)
{
code val;

scode = so;
*scode = 0;
sbyte = si;
length = maxlen;
WInitTable();

Prefix=*sbyte;
while(++pbyte<len) {
	if(((short)(val=LookUp(*(++sbyte))))==-1) {	/* Break in sequence */
		if(!WCode(Prefix)) return 0;	/* Write to buffer */
		if(!WAddPrefix()) return 0;	/* Car is new prefix */
		Prefix=*sbyte;
	} else Prefix=val;
}
if(!WCode(Prefix)) return 0;

len = pcode/sizeof(code)/8;
if(pcode&(sizeof(code)*8-1)) len++;
len *= sizeof(code);

return len;
}

/* CLEAR Code Read */
static int RClear(void)
{

bits=8;
Index=START;
Prefix=RCode();
*sbyte++=Prefix;
if(++pbyte >= length) return 0;
bits++;
return 1;
}

/* Read Lzw (decoding) */
int RLZW(code *si, byte *so,int len)
{
code val;

sbyte=so;
scode=si;
length=len;
RInitTable();

RClear();
for(;;) {
	val=RCode();
	if(val==CLEAR) {
		if(!RClear()) break;
		continue;
	}
	len=Expand(val);
	if(len<0) return len;
	if(len==0) break;
	if(Index>=(1<<bits))
		bits++;
} 
return 1;
}

long LZW_init(long mem_start)
{
PTable = (code *)mem_start;
mem_start += (LZSIZE+4)*sizeof(code);
NTable = (code *)mem_start;
mem_start += (LZSIZE+4)*sizeof(code);
CTable = (byte *)mem_start;
mem_start += (LZSIZE+4)*sizeof(byte);
memset((char *)(PTable+LZSIZE), '$', 8);
memset((char *)(NTable+LZSIZE), '$', 8);
memset(CTable+LZSIZE, '$', 4);
return mem_start;
}

int lzw_debug(void) {
	if(memcmp((char *)(PTable+LZSIZE), "$$$$$$$$", 8)) return 1;
	if(memcmp((char *)(NTable+LZSIZE), "$$$$$$$$", 8)) return 2;
	if(memcmp(CTable+LZSIZE, "$$$$", 4)) return 3;
	return 0;
}
